#!/usr/bin/env python

"""$URL: svn://svn/repos/trunk/sitereaper/setup.py $
$Id: setup.py 21296 2003-04-08 18:43:43Z akuchlin $

Setup script for sitereaper
"""

import os
from distutils import core

kw = {'name': "Sitereaper",
      'version': "0.1",
      'description': "A link checker with optional HTML validation and spell-checking",
      'author': "MEMS Exchange",
      'author_email': "quixote@mems-exchange.org",
      'url': "http://www.mems-exchange.org/software/sitereaper.html",

      'package_dir': {'sitereaper':'lib'},
      'packages': ['sitereaper'],
      'scripts': ['bin/sitereaper'],
     }

# If we're running Python 2.3, add extra information
if hasattr(core, 'setup_keywords'):
    if 'classifiers' in core.setup_keywords:
        kw['classifiers'] = ['Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Operating System :: Unix',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: MacOS :: MacOS X',
          'Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking',
          ]
    if 'download_url' in core.setup_keywords:
        kw['download_url'] = ('http://www.mems-exchange.org/software/files'
                              '/sitereaper/%s-%s.tar.gz' %
                              (kw['name'], kw['version']) )

core.setup(**kw)
