"""$URL: svn+ssh://svn/repos/trunk/sitereaper/lib/__init__.py $
$Id: __init__.py 21763 2003-05-30 20:27:55Z dbinger $

A web crawler for spellchecking and validating HTML pages.
Catches dead links at the same time, naturally.
"""

__version__ = "0.1"
